VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "TreeviewForm"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Compare Database
Option Explicit


'References: Microsoft Windows Common Controls
'If this in not in the list, browse for MSCOMCTL.OCX in the reference browser
'Normally located in C:\Windows\System32
'If you cannot find it download from microsoft
'MS scripting runtime if planning to save the sort order
'Microsoft Office 14 (or later) for command bars

Const ID_Field = "ID"
Const Parent_ID_Field = "ParentID"
Const Node_Text_Field = "NodeText"
Const Identifier_Field = "Identifier"
Const Image_Name_Field = "NodeImage"


Private WithEvents mTVW As TreeView
Attribute mTVW.VB_VarHelpID = -1
Private mQueryName As String
Private mRootParentID As String
Private mRecordSet As DAO.Recordset
Private mLoadType As lt_LoadType
Private mAllowDragDrop As Boolean
Private mTree As TreeView
Private mLoadImages As Boolean
Public Event TreeviewNodeClick(PK As Variant)
Public Event TreeviewNodeCheck(PK As Variant, Checked As Boolean)
Public Event TreeviewNodeDragDrop(DragPK As Variant, DropPK As Variant, DragNode As Node, DropNode As Node)
Public Event TreeviewNodeDblClick(PK As Variant)
Public Event RightClickOnNode(PK As Variant)
Public Event RightClickOffNode()
Public Event NodeIterated(Node As Node)
Public Event ExpandedBranch(ExpandedNode As Node)
Public Enum lt_LoadType
  lt_lightload = 1
  lt_fullload = 2
End Enum

'**************************************************************************************************************
'-------------------------------------------------------Begin Initialize Method -------------------------------
'**************************************************************************************************************
'IN ORDER TO USE THIS CLASS YOU MUST BUILD THE QUERY WITH THE MINIMUM FIELDS IN ORDER TO LOAD THE TREEVIEW
'IT WILL WORK WITH A SELFREFENCING OR YOU CAN USE A UNION QUERY TO DO THIS WITH A 1 TO MANY
'The trick is to build a query that is self referencing.  This can be done with any parent child relationship using
'a union query.  However the fields need to be aliased as follows:
'
'ID: This field is a concatenation of the Primary Key and the "identifier" field
'ParentID: This field is a concatenation of the actual foriegn key of the parent and the value of the parent record "identifier"
'NodeText: This is a concatenated field of the text to show in the node
'NodeImage: Optional if identifying the name of the image to use from the listview
'Identifier: a unique identifier to tell which table the records came from originally


' ----------------------------------------------------------------
' Procedure Name: Initialize
' Purpose: Initialize the Class using default parameters
' Procedure Kind: Sub
' Procedure Access: Public
' Parameter objTreeView (TreeView): Reference to the Treeview on the form
' Parameter QueryName (String): This is a specific query format using alias that allows you
'    to turn any query into a Treeview
' Parameter parentStartID (String): Need to look in the query and determine starting ID of root level.
'     Normally this is just the Identifier.
' Author: MajP
' Date: 3/5/2020
' ----------------------------------------------------------------
Public Sub Initialize(objTreeView As TreeView, QueryName As String, parentStartID As String, Optional LoadImages As Boolean = False, Optional LoadType As lt_LoadType = lt_lightload, Optional AllowDragDrop As Boolean = False)
   
  On Error GoTo ErrHandler

  Set mTVW = objTreeView
  mTVW.Nodes.Clear
  mQueryName = QueryName
  mRootParentID = parentStartID
  mLoadImages = LoadImages
  mLoadType = LoadType
  Me.AllowDragDrop = AllowDragDrop
  Set mRecordSet = CurrentDb.OpenRecordset(QueryName, dbOpenDynaset)
  Call addRootNodes(parentStartID)
  Exit Sub
ErrHandler:
      MsgBox ("Error: " & Err.Number _
            & " " & Err.Description _
            & " " & Err.Source)
   Resume Next
End Sub

'**************************************************************************************************************
'-------------------------------------------------------Procedures for Loading Tree View ----------------------
'**************************************************************************************************************


Public Sub addRootNodes(StartID As Variant)
  Dim strCriteria As String
  Dim bk As String
  Dim currentID As Variant
  Dim currentText As String
  Dim currentNode As Node
  Dim ParentNode As Node
  Dim currentIdentifier As String
  Dim currentImage As String
  strCriteria = Parent_ID_Field & " = '" & StartID & "'"
  Me.Recordset.FindFirst (strCriteria)
  If Me.Recordset.NoMatch Then
    MsgBox "There is no record with a Parent ID of " & StartID
  End If
  Do Until Me.Recordset.NoMatch
   currentID = Me.Recordset.Fields(ID_Field)
   currentText = Me.Recordset.Fields(Node_Text_Field)
   currentIdentifier = Me.Recordset.Fields(Identifier_Field)
   Set currentNode = mTVW.Nodes.Add(, , currentID, currentText)
   If mLoadImages Then
     currentImage = Nz(Me.Recordset.Fields(Image_Name_Field), "")
     If currentImage <> "" Then currentNode.Image = currentImage
   End If
   currentNode.Tag = currentIdentifier
   bk = Me.Recordset.Bookmark
   'Determine if Light or Full Load
   If mLoadType = lt_fullload Then
     Call addBranch(currentID, currentIdentifier)
   Else
     Call addLiteBranch(currentID, currentIdentifier)
   End If
   Me.Recordset.Bookmark = bk
   Me.Recordset.FindNext (strCriteria)
  Loop
End Sub

Public Sub addBranch(parentIDValue As Variant, identifier)
  On Error GoTo errLable
  Dim strCriteria As String
  Dim bk As String
  Dim currentID As Variant
  Dim currentText As String
  Dim currentNode As Node
  Dim ParentNode As Node
  Dim strKey As String
  Dim currentImage As String
  Dim rsTree As Recordset
  Set rsTree = Me.Recordset
  strCriteria = Parent_ID_Field & " = '" & parentIDValue & "'"
  
  'Since it is not a root node must determine the parent node.  All node keys are "ID " & the primary key
  Set ParentNode = Me.Nodes(parentIDValue)
  'MsgBox strCriteria
  rsTree.FindFirst (strCriteria)
  'MsgBox strCriteria
  Do Until rsTree.NoMatch
    currentID = rsTree.Fields(ID_Field)
    currentText = rsTree.Fields(Node_Text_Field)
    identifier = rsTree.Fields(Identifier_Field)
   Set currentNode = Me.Nodes.Add(ParentNode, tvwChild, currentID, currentText)
   currentNode.Tag = identifier
   If mLoadImages Then
     currentImage = Nz(rsTree.Fields(Image_Name_Field), "")
     If currentImage <> "" Then currentNode.Image = currentImage
   End If
   bk = rsTree.Bookmark
   Call addBranch(currentID, identifier)
    rsTree.Bookmark = bk
    rsTree.FindNext (strCriteria)
  Loop
  Exit Sub
errLable:
  MsgBox Err.Number & " " & Err.Description & " In addBranch"
  If MsgBox("Do you want to exit the loop?", vbYesNo, "Error In Loop") = vbYes Then
     Exit Sub
   Else
     Resume Next
   End If
End Sub

Public Sub addLiteBranch(parentIDValue As Variant, identifier)
  On Error GoTo errLable
  Dim strCriteria As String
  Dim bk As String
  Dim currentID As Variant
  Dim currentText As String
  Dim currentNode As Node
  Dim currentImage As String
  Dim ParentNode As Node
  Dim strKey As String
  Dim rsTree As DAO.Recordset
  Set rsTree = Me.Recordset
  strCriteria = Parent_ID_Field & " = '" & parentIDValue & "'"
  
  'Since it is not a root node must determine the parent node.  All node keys are "ID " & the primary key
  Set ParentNode = Me.Nodes(parentIDValue)
  rsTree.FindFirst (strCriteria)
  'Only going to get the top node
  If Not rsTree.NoMatch Then
    currentID = rsTree.Fields(ID_Field)
    currentText = rsTree.Fields(Node_Text_Field)
    identifier = rsTree.Fields(Identifier_Field)
   Set currentNode = Me.Nodes.Add(ParentNode, tvwChild, currentID, currentText)
   currentNode.Tag = identifier
   If mLoadImages Then
     currentImage = Nz(rsTree.Fields(Image_Name_Field), "")
     If currentImage <> "" Then currentNode.Image = currentImage
   End If
   bk = rsTree.Bookmark
   Call addLiteBranch(currentID, identifier)
    rsTree.Bookmark = bk
    rsTree.FindNext (strCriteria)
  End If
  Exit Sub
errLable:
  MsgBox Err.Number & " " & Err.Description & " In addBranch"
  If MsgBox("Do you want to exit the loop?", vbYesNo, "Error In Loop") = vbYes Then
     Exit Sub
   Else
     Resume Next
   End If
End Sub

Private Sub mTVW_Expand(ByVal Node As MSComctlLib.Node)
 'Flush out any remaining nodes in this expanded level and liteload child level
 'This procedure is associated with a light load. Occurs when you expand the branch
  On Error GoTo errlbl
  Dim strCriteria As String
  Dim bk As String
  Dim currentID As Variant
  Dim currentText As String
  Dim currentNode As Node
  Dim ParentNode As Node
  Dim currentIdentifier As String
  Dim currentImage As String
  Dim rsTree As DAO.Recordset
  'Debug.Print Node.Text & " " & "Selected Node"
  If Node.Children > 1 Then Exit Sub ' node has already been expanded
  Set rsTree = Me.Recordset
  strCriteria = Parent_ID_Field & " = '" & Node.key & "'"
  rsTree.FindFirst (strCriteria)
  Debug.Print strCriteria
  If rsTree.NoMatch Then
  '  MsgBox "There is no record with a Parent ID of " & Node.key
  End If
  
  Do Until rsTree.NoMatch
   currentID = rsTree.Fields(ID_Field)
   Debug.Print currentID & " current"
   currentText = rsTree.Fields(Node_Text_Field)
   currentIdentifier = rsTree.Fields(Identifier_Field)
   If Node.Child.key = currentID Then
     Node.Child.Tag = currentIdentifier
     RaiseEvent ExpandedBranch(Node.Child)
   Else
     'If you are light loaded and drag drop then you will get a duplicated key error
     Set currentNode = mTVW.Nodes.Add(Node.key, tvwChild, currentID, currentText)
     currentNode.Tag = currentIdentifier
     If mLoadImages Then
       currentImage = Nz(rsTree.Fields(Image_Name_Field), "")
       If currentImage <> "" Then currentNode.Image = currentImage
     End If
     RaiseEvent ExpandedBranch(currentNode)
     bk = rsTree.Bookmark
     Call addLiteBranch(currentID, currentIdentifier)
   End If
   rsTree.Bookmark = bk
   rsTree.FindNext (strCriteria)
  Loop
  Exit Sub
errlbl:
  If Not Err.Number = 35602 Then 'this is a duplicate key caused by drag drop and light loaded
    MsgBox Err.Number & " " & Err.Description & " Key: " & currentID & " In Treeview Expand."
  End If
End Sub

Public Sub reloadTree()
  If mQueryName <> "" Then
    Me.ClearTree
    Me.Initialize mTVW, mQueryName, mRootParentID, mLoadImages, mLoadType
  End If
End Sub

'**************************************************************************************************************
'-------------------------------------------------------Class Properties---------------------------------------
'**************************************************************************************************************

Public Property Get AllowDragDrop() As Boolean
   AllowDragDrop = mAllowDragDrop
End Property

Public Property Let AllowDragDrop(yesNo As Boolean)
  mAllowDragDrop = yesNo
  If yesNo Then
    mTVW.OLEDragMode = 1
    mTVW.OLEDropMode = 1
  Else
    mTVW.OLEDragMode = 0
    mTVW.OLEDropMode = 0
  End If
End Property

Public Property Get Nodes() As Nodes
  Set Nodes = Me.TreeView.Nodes
End Property

Public Property Get Recordset() As DAO.Recordset
  Set Recordset = mRecordSet
End Property

Public Property Get Selectednode() As Node
   Set Selectednode = mTVW.SelectedItem
End Property

Public Property Get SelectedNodeIdentifier() As String
  If Not Me.Selectednode Is Nothing Then
    SelectedNodeIdentifier = Me.Selectednode.Tag
  Else
    MsgBox "No Node selected."
  End If
End Property

Public Property Get SelectedNodeLevel() As Integer
   Dim myNode As Node
   Dim intCount As Integer
   Set myNode = mTVW.SelectedItem
   If Not myNode Is Nothing Then
      intCount = 1
      Do Until FindRoot(myNode, mTVW).key = myNode.key
         Set myNode = myNode.Parent
         intCount = intCount + 1
      Loop
      SelectedNodeLevel = intCount
   End If
End Property

Public Property Get SelectedNodePK() As Variant
  If Not Me.Selectednode Is Nothing Then
    SelectedNodePK = Replace(Me.Selectednode.key, Me.SelectedNodeIdentifier, "")
  Else
    MsgBox "No Node selected"
  End If
End Property

Public Property Get TreeView() As TreeView
  Set TreeView = mTVW
End Property
'**************************************************************************************************************
'----------------------------------------------------End Class Properties--------------------------------------------------
'**************************************************************************************************************

'**************************************************************************************************************
'--------------------------------------------Class Methods   --------------------------------------------------
'**************************************************************************************************************

Public Sub ClearTree()
  Me.Nodes.Clear
End Sub

Public Sub CollapseBranch()
    Dim theNode As Node
    Set theNode = mTVW.SelectedItem
    Do While Not theNode Is Nothing
      theNode.Expanded = False
      Set theNode = theNode.Parent
    Loop
End Sub

Public Sub CollapseTree()
On Error GoTo Err_CollapseTree
    Dim myNode As Node
    For Each myNode In Me.Nodes
      myNode.Expanded = False
    Next myNode
Exit_CollapseTree:
    Exit Sub
Err_CollapseTree:
    MsgBox Err.Description & " In Collapse Tree."
    Resume Exit_CollapseTree
End Sub

Public Sub DeleteNode(strIndex As String)
  On Error Resume Next
     Me.Nodes.Remove (strIndex)
End Sub

Public Sub deleteSelectedNode()
  If Not Me.Selectednode Is Nothing Then
    Me.Nodes.Remove (Selectednode.Index)
  Else
    MsgBox "No Node selected."
  End If
End Sub

Public Sub ExpandTree()
   On Error GoTo Err_ExpandTree
   'The recursion had to be done to handle light load
   'If not you would just expand each node.
   Dim SN As Node
   Set SN = Me.Selectednode
   Dim tempNode As Node
   Set tempNode = Me.TreeView.Nodes(1)
   Do
     expandChildren tempNode
     Set tempNode = tempNode.Next
   Loop Until tempNode Is Nothing
Exit_ExpandTree:
      Exit Sub
Err_ExpandTree:
      If Err.Number = 35606 Then
        Resume Next
      Else
        MsgBox Err.Number & " " & Err.Description & " In Expand Tree."
        Resume Exit_ExpandTree
      End If
End Sub

Public Sub ExpandBranch()
   Dim Selectednode As Node
   If Not mTVW.SelectedItem Is Nothing Then
     Set Selectednode = mTVW.SelectedItem
     Call expandChildren(Selectednode)
   End If
End Sub

Private Sub expandChildren(theNode As Node)
   'This one is not that simple
   'It requires a function to return the child nodes since this is not a
   'property of a Node.  The "children" property returns the count of the children
   ' This then uses a recursive call to span the branch.
   Dim myNode As Node
   theNode.Expanded = True
   If mLoadType = lt_lightload Then
      mTVW_Expand theNode
   End If
   For Each myNode In childNodes(theNode)
     If Not myNode.Child Is Nothing Then
       expandChildren myNode
     End If
   Next myNode
   Set myNode = Nothing
End Sub

Public Sub IterateBranch(ParentNode As Node)
  On Error GoTo errLable
  'MsgBox strCriteria
  Dim childNodes As Collection
  Dim childNode As Node

  
  If ParentNode Is Nothing Then
    MsgBox "No Node Passed in Iterate Branch. ", vbInformation
    Exit Sub
  End If
   Set childNodes = Me.childNodes(ParentNode)
  
  For Each childNode In childNodes
    Debug.Print "in loop"
    RaiseEvent NodeIterated(childNode)
   Call IterateBranch(childNode)
  Next childNode
  Exit Sub
errLable:
  MsgBox Err.Number & " " & Err.Description & " In addBranch"
  If MsgBox("Do you want to exit the loop?", vbYesNo, "Error In Loop") = vbYes Then
     Exit Sub
   Else
     Resume Next
   End If
End Sub

Public Sub MoveDown()
  Dim FollowingNode As Node
  Dim nodeTomove As Node
  Dim nodeToMoveKey As String
  Dim nodeToMoveText As String
  Dim nodeToMoveImage As String
  Dim Siblings As Collection
  Dim I As Integer
  Dim j As Integer
  Dim prevKey As String
  Dim prevText As String
  Dim prevImage As String
  Dim NodeToMoveChildren As Collection
  Dim FollowingNodeChildren As Collection
  
  Set nodeTomove = Me.Selectednode
  If nodeTomove Is Nothing Then
    MsgBox "No node selected", vbInformation
    Exit Sub
  End If
  nodeToMoveKey = nodeTomove.key
  nodeToMoveText = nodeTomove.Text
  nodeToMoveImage = nodeTomove.Image
  'Set FollowingNode = nodeTomove.Previous
   
 If Not nodeTomove.Next Is Nothing Then
      Set FollowingNode = nodeTomove.Next
      If Not FollowingNode Is Nothing Then
        FollowingNode.Selected = True
        prevKey = FollowingNode.key
        prevText = FollowingNode.Text
        If FollowingNode.Image <> "" Then prevImage = FollowingNode.Image
        FollowingNode.key = ""
        nodeTomove.key = ""
        FollowingNode.Text = nodeToMoveText
        FollowingNode.key = nodeToMoveKey
        If nodeToMoveImage <> "" Then FollowingNode.Image = nodeToMoveImage
        nodeTomove.Text = prevText
        nodeTomove.key = prevKey
        If prevImage <> "" Then nodeTomove.Image = prevImage
        
        Set NodeToMoveChildren = Me.childNodes(nodeTomove)
        Set FollowingNodeChildren = Me.childNodes(FollowingNode)
         
        For j = 1 To NodeToMoveChildren.Count
            Set NodeToMoveChildren(j).Parent = FollowingNode
        Next j
        For j = 1 To FollowingNodeChildren.Count
          Set FollowingNodeChildren(j).Parent = nodeTomove
        Next j
      End If
      Set Me.TreeView.DropHighlight = FollowingNode
 Else
   MsgBox "You are at the bottom of your level.", vbInformation
 End If
End Sub

'Constants used to define relationship of node when adding
'CONSTANT / DESCRIPTION
'tvwFirst / Add as the first node at the level of the relative.
'tvwLast / Add as the last node at the level of the relative.
'tvwNext / Add after immediately following a specified node.
'tvwPrevious / Add after immediately preceding a specified node.
'tvwChild / Add as a child to the specified node.

Public Sub MoveUp()
  Dim prevNode As Node
  Dim nodeTomove As Node
  Dim nodeToMoveKey As String
  Dim nodeToMoveText As String
  Dim nodeToMoveImage As String
  Dim Siblings As Collection
  Dim I As Integer
  Dim j As Integer
  Dim prevKey As String
  Dim prevText As String
  Dim prevImage As String
  Dim NodeToMoveChildren As Collection
  Dim PrevNodeChildren As Collection
  
  Set nodeTomove = Me.Selectednode
  If nodeTomove Is Nothing Then
    MsgBox "No node selected", vbInformation
    Exit Sub
  End If
  nodeToMoveKey = nodeTomove.key
  nodeToMoveText = nodeTomove.Text
  nodeToMoveImage = nodeTomove.Image
  If Not nodeTomove.Previous Is Nothing Then
      Set prevNode = nodeTomove.Previous
      'MsgBox prevNode.Text
      If Not prevNode Is Nothing Then
        prevNode.Selected = True
        prevKey = prevNode.key
        prevText = prevNode.Text
        prevImage = prevNode.Image
        prevNode.key = ""
        nodeTomove.key = ""
        prevNode.Text = nodeToMoveText
        prevNode.key = nodeToMoveKey
        If prevNode.Image <> "" Then prevNode.Image = nodeToMoveImage
        nodeTomove.Text = prevText
        nodeTomove.key = prevKey
        If prevImage <> "" Then nodeTomove.Image = prevImage
        Debug.Print "move children "
        Set NodeToMoveChildren = Me.childNodes(nodeTomove)
        Set PrevNodeChildren = Me.childNodes(prevNode)
        For j = 1 To NodeToMoveChildren.Count
          Set NodeToMoveChildren(j).Parent = prevNode
        Next j
        For j = 1 To PrevNodeChildren.Count
          Set PrevNodeChildren(j).Parent = nodeTomove
        Next j
        Set Me.TreeView.DropHighlight = prevNode
      End If
   Else
     MsgBox "You are at the top of your level.", vbInformation
 End If
  
End Sub

Public Sub rollCall()
  Dim theNode As Node
  For Each theNode In Me.Nodes
    Debug.Print theNode.key
  Next theNode
End Sub

Public Sub SelectNode(nodeID As String)
  mTVW.Nodes(nodeID).Selected = True
End Sub


'**************************************************************************************************************
'-------------------------------------------------------Begin Class Functions ---------------------------------
'**************************************************************************************************************

Public Function GetDescendants(theNode As Node, Optional theDescendants As Collection = Nothing) As Collection
  'This gets all children and sub children.  These are a nodes Descendants
  Dim nod As Node
  If theDescendants Is Nothing Then Set theDescendants = New Collection
  For Each nod In childNodes(theNode)
    theDescendants.Add nod
    If Not nod.Child Is Nothing Then
      GetDescendants nod, theDescendants
    End If
  Next nod
  Set GetDescendants = theDescendants
End Function

Private Function FindRoot(ByVal nodX As Node, TVW As TreeView) As Node
   Dim n As Integer
   On Error GoTo ErrHandler
    n = nodX.Index
    While n <> nodX.Root.Index
        n = TVW.Nodes(n).Parent.Index
   Wend
ErrHandler:
       Set FindRoot = TVW.Nodes(n)
End Function

Public Function GetLevelIDs(Optional StartLevel As Integer = 1, Optional Modifier As String = "L") As Dictionary
  'Returns a dictionary of PKs and Levels
  Dim col As New Collection
  'need scripting runtime
  Dim NLI_Dict As Dictionary
  Dim theNode As Node
  Dim topLevelSort As Integer
  Dim levelSort As Integer
  Dim LevelID As String
  Dim oldParentID As String
  Dim currentParentID As String
  Dim ParentLevelID As String
  Set NLI_Dict = CreateObject("Scripting.Dictionary")
  NLI_Dict.RemoveAll
  topLevelSort = StartLevel
  levelSort = StartLevel
  Me.TreeView.Refresh
  For Each theNode In Me.Nodes
    If theNode.Parent Is Nothing Then
      LevelID = Modifier & topLevelSort
      topLevelSort = topLevelSort + 1
       If Not NLI_Dict.Exists(Me.getNodePK(theNode)) Then
          NLI_Dict.Add Me.getNodePK(theNode), LevelID
        Else
          Debug.Print "error Exists" & (Me.getNodePK(theNode))
        End If
    Else
        levelSort = GetNodeLevelSort(theNode)
        LevelID = NLI_Dict.Item(Me.getNodePK(theNode.Parent))
        LevelID = LevelID & "." & levelSort
        Debug.Print Me.getNodePK(theNode) & " " & LevelID
        If Not NLI_Dict.Exists(Me.getNodePK(theNode)) Then
          NLI_Dict.Add Me.getNodePK(theNode), LevelID
        Else
          Debug.Print "error Exists" & (Me.getNodePK(theNode))
        End If
    End If
    
  Next theNode
  Set GetLevelIDs = NLI_Dict
End Function

Public Function getNode(strKey As String) As Node
  On Error GoTo errLable
  Set getNode = Me.Nodes(strKey)
  Exit Function
errLable:
  If Err.Number = 35601 Then
    MsgBox "No Node exists with a key of " & strKey
  Else
    MsgBox Err.Number & " " & Err.Description
  End If
End Function

Public Function childNodes(ByVal TheParentNode As Node) As Collection
  Dim intCounter As Integer
  Set childNodes = New Collection
  Dim theNode As Node
  
  If Not TheParentNode.Child Is Nothing Then
    Set theNode = TheParentNode.Child
  Else
    Exit Function
  End If
  
  Do While Not theNode Is Nothing
    childNodes.Add theNode
    Set theNode = theNode.Next
  Loop
End Function

Public Function getNodeIdentifier(theNode As Node) As String
  getNodeIdentifier = theNode.Tag
End Function

Public Function GetNodeLevel(myNode As Node) As Integer
   'saving the PK as the key and the level in the value
   Dim intCount As Integer
   If Not myNode Is Nothing Then
      intCount = 1
      Do Until FindRoot(myNode, mTVW).key = myNode.key
         Set myNode = myNode.Parent
         intCount = intCount + 1
      Loop
      GetNodeLevel = intCount
   End If
End Function

Public Function GetNodeLevelSort(theNode As Node) As Integer
   'This works on the root nodes but I do not know why
   Dim thechildren As Collection
   Dim I As Integer
   If Not theNode.Parent Is Nothing Then
     Set thechildren = Me.childNodes(theNode.Parent)
     'Debug.Print thechildren.Count
     For I = 1 To thechildren.Count
       If thechildren(I).key = theNode.key Then
        GetNodeLevelSort = I
        Exit Function
      End If
     Next I
   Else
     MsgBox "Not been updated to handle root nodes", vbInformation
   End If
End Function

Public Function getNodePK(theNode As Node) As String
  getNodePK = Replace(theNode.key, theNode.Tag, "")
End Function

Public Function GetSortOrder() As Collection
  'used to get the sort order to populate the table
  'Used in single table hierrarchy
  Dim col As New Collection
  Dim theNode As Node
  For Each theNode In Me.Nodes
    'Debug.Print theNode.Key
    col.Add (Me.getNodePK(theNode))
  Next theNode
  Set GetSortOrder = col
End Function

'**************************************************************************************************************
'---------------------------------------------Treeview Trapped Events------------------------------------------
'**************************************************************************************************************

Private Sub mTVW_DblClick()
   If Not Selectednode Is Nothing Then
     RaiseEvent TreeviewNodeDblClick(Me.getNodePK(Me.Selectednode))
   End If
End Sub

Private Sub mTVW_MouseUp(ByVal Button As Integer, ByVal Shift As Integer, ByVal x As stdole.OLE_XPOS_PIXELS, ByVal y As stdole.OLE_YPOS_PIXELS)
  Dim NodeHit As Node
  If Button = 2 Then
    Set NodeHit = mTVW.HitTest(x, y)
    If Not NodeHit Is Nothing Then
      RaiseEvent RightClickOnNode(Me.getNodePK(NodeHit))
    Else
      RaiseEvent RightClickOffNode
    End If
  End If
End Sub

Private Sub mTVW_NodeCheck(ByVal Selectednode As MSComctlLib.Node)
  If Selectednode.Checked = True Then
    Selectednode.Bold = True
    RaiseEvent TreeviewNodeCheck(Me.SelectedNodePK, True)
   Else
    Selectednode.Bold = False
    RaiseEvent TreeviewNodeCheck(Me.SelectedNodePK, False)
  End If
End Sub

Private Sub mTVW_NodeClick(ByVal Selectednode As MSComctlLib.Node)
 'Here is where you can fire an event off
 'of a click. Ensure you have a procedure in a common module called commonNodeClick.
On Error GoTo Err_NodeClick
  RaiseEvent TreeviewNodeClick(Me.getNodePK(Selectednode))
Exit_NodeClick:
    Exit Sub
Err_NodeClick:
    MsgBox Err.Description
    Resume Exit_NodeClick
End Sub

Private Sub mTVW_OLEDragDrop(Data As MSComctlLib.DataObject, Effect As Long, Button As Integer, Shift As Integer, x As Single, y As Single)
    On Error GoTo ErrxTree_OLEDragDrop
    Dim strKey As String
    Dim strText As String
    Dim nodNew As Node
    Dim nodDragged As Node
    Dim nodTemp As Node
    'If nothing is selected for drag, do nothing.
    If Not Me.AllowDragDrop Then
      MsgBox "You Have not allowed Drag Drop. You have to set AllowDragDrop to true to use this feature"
      Exit Sub
    End If
    If Not Me.Selectednode Is Nothing Then
       'Reference the selected node as the one being dragged.
        Set nodDragged = Me.Selectednode
        strKey = nodDragged.key
        strText = nodDragged.Text
        'If the node was dragged to an empty space, update the
         If mTVW.DropHighlight Is Nothing Then
            'Save the key and the text to use when you re-add the node.
            'Delete the current node
             'Add this node as a root node.
              Me.Nodes.Remove nodDragged.Index
              Set nodNew = Me.Nodes.Add(, , strKey, strText)
              nodNew.Tag = nodDragged.Tag
              If mLoadImages Then
                nodNew.Image = DLookup("nodeImage", mQueryName, "ID ='" & strKey & "'")
               End If
              'Add all the child nodes for this node
              addBranch strKey, nodNew.Tag
              Set nodTemp = nodNew
            'If you are not dropping the node on itself.
            RaiseEvent TreeviewNodeDragDrop(Me.getNodePK(nodNew), Null, nodNew, nodNew)
        ElseIf nodDragged.Index <> mTVW.DropHighlight.Index Then
            'Set the drop target as the selected node's parent.
             Set nodDragged.Parent = mTVW.DropHighlight
             Set nodTemp = nodDragged
             RaiseEvent TreeviewNodeDragDrop(Me.getNodePK(nodDragged), Me.getNodePK(nodDragged.Parent), nodDragged, nodDragged.Parent)
        End If

    End If

    'Deselect the node
    Set nodDragged = Nothing

    'Unhighlight the nodes.
    Set mTVW.DropHighlight = Nothing
 
ExitxTree_OLEDragDrop:
    Exit Sub
ErrxTree_OLEDragDrop:
    'If you create a circular branch.
    If Err.Number = 35614 Then
        MsgBox "You created a cirular reference.", _
            vbCritical, "Move Cancelled"
    Else
        MsgBox Err.Description
    End If
    Resume ExitxTree_OLEDragDrop
End Sub

Private Sub mTVW_OLEDragOver(Data As MSComctlLib.DataObject, Effect As Long, Button As Integer, Shift As Integer, x As Single, y As Single, State As Integer)

'====================================================================
'Use the OLEDragOver event of the TreeView control to select the
'node to drag, and to highlight the target nodes where the drop will
'occur when you release the mouse. This procedure sets the selected
'node to drag once. After that, if a node is already selected, the
'procedure assumes it was set during an earlier call in the dragging
'process and it does not reset it.  The second half of this procedure
'highlights the node you are dragging over.
'====================================================================
    'If no node is selected, select the first node you dragged over.
    If Me.Selectednode Is Nothing Then
        Set mTVW.SelectedItem = mTVW.HitTest(x, y)
    End If

    'Highlight the node being dragged over as a potential drop target.
    Set mTVW.DropHighlight = mTVW.HitTest(x, y)
End Sub

Private Sub mTVW_OLEStartDrag(Data As MSComctlLib.DataObject, AllowedEffects As Long)
'==================================================================
'This procedure in the OLEStartDrag event of the TreeView control
'clears the selected node so you can choose a new one.
'==================================================================
    mTVW.SelectedItem = Nothing
    'MsgBox "start"
End Sub
'**************************************************************************************************************
'-------------------------------------------------------End Tree view Events --------------------------------
'**************************************************************************************************************

Private Sub Class_Terminate()
 Set mTVW = Nothing
End Sub

